/* $Id: options.c,v 1.2 96/10/31 23:34:28 lee Exp Locker: lee $
 *
 * Parse command-line options.
 *
 */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "api.h"
#define NO_X_INCLUDES
#include "fonttool.h"
#include "options.h"

typedef enum {
    InternalError,
    UnknownOption,
    InvalidArgument,
    MissingArgument,
    OK
} t_optionStatus;

#define OPTFLAG_HAS_BEEN_SEEN	001
#define OPTFLAG_HANDLED		002
#define OPTFLAG_USED_DEFAULT	004
#define OPTFLAG_UNDOCUMENTED	010

#define DEFAULT_CONSTRAINT "-*-*-*-*-*-*-0-0-0-0-*-0-*-*"

typedef struct s_option {
    char *names; /* synonyms separated by | */
    char *description;
    int argumentCount;
    t_optionStatus (* handler)(
#ifdef HAVE_PROTO
	struct s_option *optp,
	char *optionGiven,
	char *arg1, /* optional, NULL pointer if not set */
	char *arg2
#endif
    );
    char *ptr;
    char *defaultValue;
    unsigned char flags;
} t_option;

extern char *fieldNames[];
extern char *ConstrainToTheseFields[];

#define SET_ALWAYS 01
#define SET_IFNULL 02

static void
setConstraint(field, value, when)
    int field;
    char *value;
    int when;
{
    if (when == SET_ALWAYS) {
	if (ConstrainToTheseFields[field]) {
	    (void) free(ConstrainToTheseFields[field]);
	    ConstrainToTheseFields[field] = 0;
	}
    }

    if (!ConstrainToTheseFields[field]) {
	register char *p;

	/* save a copy of the string so we can always free it */
	ConstrainToTheseFields[field] = strdup(value);

	/* convert to lower case */
	for (p = ConstrainToTheseFields[field]; *p; p++) {
	    if (isupper(*p)) {
		*p = tolower(*p);
	    }
	}
    }
}

extern int fprintf();

static t_optionStatus
setField(optp, optionGiven, arg1, arg2)
    struct s_option *optp;
    char *optionGiven;
    char *arg1;
    char *arg2;
{
    int field;

    if (optp->argumentCount != 1 || !arg1 || !*arg1) {
	fprintf(stderr,
	    "%s: %s must be followed by a value\n",
	    progname, optionGiven
	);
	exit(2);
    }

    /* turn -family into FIELD_FAMILY: */
    field = fileNameToNumber(&optionGiven[1]);
    if (field < 0) {
	int i;
	fprintf(stderr,
	    "%s: %s not a recognised field; choices are:\n",
	    progname, optionGiven
	);

	for (i = 0; fieldNames[i]; i++) {
	    fprintf(stderr, "%s:    -%s \"string\"\n",
		progname, fieldNames[i]
	    );
	}
	fprintf(stderr, "%s -help gives more usage information.\n", progname);
	exit(2);
    }

    setConstraint(field, arg1, SET_ALWAYS);
    return OK;
}

static t_optionStatus
setPattern()
{
}

static t_optionStatus
setEncoding()
{
}

static t_optionStatus
setRes()
{
}

static t_optionStatus
setScaled()
{
    /* Constrain the font pattern so that it only matches
     * scaled fonts -- i.e. ones with a zero for size, dpi and width
     */

    setConstraint(FIELD_PIXELSZ, "0", SET_IFNULL);
    setConstraint(FIELD_POINTSIZE, "0", SET_IFNULL);
    setConstraint(FIELD_RESX, "0", SET_IFNULL);
    setConstraint(FIELD_RESY, "0", SET_IFNULL);
    setConstraint(FIELD_AVERAGEWIDTH, "0", SET_IFNULL);
    return OK;
}

static void
AllowNonScaled(field)
    int field;
{
    /* If a constraint is to "0" or is null, set it to "*",
     * so that it matches any font and not just base scalable font names
     */

    if (ConstrainToTheseFields[field]) {
	if (strcmp(ConstrainToTheseFields[field], "0") == 0) {
	    free(ConstrainToTheseFields[field]);
	    ConstrainToTheseFields[field] = (char *) 0;
	}
    }
    setConstraint(field, "*", SET_IFNULL);
}

static t_optionStatus
setIncludeAll()
{
    /* Allow any font to match, and not just scaled ones */

    AllowNonScaled(FIELD_PIXELSZ);
    AllowNonScaled(FIELD_POINTSIZE);
    AllowNonScaled(FIELD_RESX);
    AllowNonScaled(FIELD_RESY);
    AllowNonScaled(FIELD_AVERAGEWIDTH);
    return OK;
}

static t_option options[] = {
    {
	(char *) "-prefix|-fontprefix|-pattern",
	(char *) "initial pattern to match",
	0,
	setPattern,
	0,
	0,
	0
    },
    {
	(char *) "-foundry",
	(char *) "match only fonts made at the given foundry (Adobe, Monotype...)",
	1,
	setField,
	(char *) FIELD_FOUNDRY,
	(char *) "*",
	0
    },
    {
	(char *) "-family",
	(char *) "match fonts in the given font family (Times, Palatino...)",
	1,
	setField,
	(char *) FIELD_FAMILY,
	(char *) "*",
	0
    },
    {
	(char *) "-weight",
	(char *) "initial font weight (bold, mediumm, light...)",
	1,
	setField,
	(char *) FIELD_WEIGHT,
	(char *) "*",
	0
    },
    {
	(char *) "-slant",
	(char *) "initial font slant (r, i or o)",
	1,
	setField,
	(char *) FIELD_SLANT,
	(char *) "*",
	0
    },
    {
	(char *) "-setwidth",
	(char *) "initial font set width (condensed, normal...)",
	1,
	setField,
	(char *) FIELD_SETWIDTH,
	(char *) "*",
	0
    },
    {
	(char *) "-addedstyle|-adstyle|-addstyle|-style",
	(char *) "inital font additional style (sans, outline...)",
	1,
	setField,
	(char *) FIELD_ADDEDSTYLE,
	(char *) "*",
	0
    },
    {
	(char *) "-pixelsz|-pixelsize",
	(char *) "initial font height in pixels (e.g. 15)",
	1,
	setField,
	(char *) FIELD_PIXELSZ,
	(char *) "0",
	0
    },
    {
	(char *) "-pointsize|-size",
	(char *) "inital font size in decipoints (e.g. 120)",
	1,
	setField,
	(char *) FIELD_POINTSIZE,
	(char *) "0",
	0
    },
    {
	(char *) "-resx|-xres|-dpix|-xpi",
	(char *) "inital x resolution for display (75 or 100 usually)",
	1,
	setField,
	(char *) FIELD_RESX,
	(char *) "0",
	0
    },
    {
	(char *) "-resy|-yres|-dpiy|-ypi",
	(char *) "inital y resolution for display (75 or 100 usually)",
	1,
	setField,
	(char *) FIELD_RESY,
	(char *) "0",
	0
    },
    {
	(char *) "-res|-dpi",
	(char *) "inital x and y resolution for display (75 or 100 usually)",
	1,
	setRes,
	(char *) 0,
	(char *) "0",
	0
    },
    {
	(char *) "-space|-spacing",
	(char *) "inital font spacing: m (monospaced), p (proportional) or c (cell)",
	1,
	setField,
	(char *) FIELD_SPACE,
	(char *) "*",
	0
    },
    {
	(char *) "-averagewidth|-avgwidth",
	(char *) "inital font average character width (0 or * for unspecified)",
	1,
	setField,
	(char *) FIELD_AVERAGEWIDTH,
	(char *) "0",
	0
    },
    {
	(char *) "-registry",
	(char *) " initial font character set registry (adobe, ISO8859...)",
	1,
	setField,
	(char *) FIELD_REGISTRY,
	(char *) "*",
	0
    },
    {
	(char *) "-encoding|-enc",
	(char *) " initial font character set (symbol, fontspecific, 1, 2...)",
	1,
	setField,
	(char *) FIELD_ENCODING,
	(char *) "*",
	0
    },
    {
	(char *) "-charset|-cset",
	(char *) " initial font full character set (adobe-fontspecific, ISO8859-1...)",
	1,
	setEncoding,
	(char *) 0,
	(char *) 0,
	0
    },
    {
	(char *) "-scaled",
	(char *) "set size and resolution fields to list only scalable font names",
	0,
	setScaled,
	(char *) "-*-*-*-*-*-*-0-0-0-0-*-0-*-*",
	(char *) 0,
	0
    },
    {
	(char *) "-all",
	(char *) "set size and resolution fields to list all XLFD font names",
	0,
	setIncludeAll,
	(char *) "-*-*-*-*-*-*-0-0-0-0-*-0-*-*",
	(char *) 0,
	0
    },
    {
	0,
    }
};

static int
foundInLine(option, optionList)
    char *option;
    char *optionList;
{
    char *p, *q = optionList;

    while (*q) {
	p = option;
	while (*p && *p == *q) {
	    if (*q == '|') {
		/* the user included | in an option!  Sneaky user! */
		break;
	    }
	    p++;
	    q++;
	}

	if (!*p) {
	    if (!*q || *q == '|') {
		return 1; /* found it */
	    }
	}

	while (*q && *q != '|') {
	    q++;
	}
	if (*q == '|') {
	    q++;
	}
    }

    return 0; /* not found */
}

static int
handleOneOption(argcp, ppp)
    int *argcp;
    char ***ppp;
{
    t_option *optp;

    for (optp = options; optp->names; optp++) {
	if (foundInLine(**ppp, optp->names)) {
	    int result;

	    /* we've identified the option.
	     * now pass the given argument to the function to handle
	     */
	    if (optp->argumentCount > *argcp - 1) {
		if (optp->argumentCount > 1) {
		    fprintf(stderr,
			"%s: option %s must be followed by %d arguments\n",
			progname, **ppp, optp->argumentCount
		    );
		} else {
		    fprintf(stderr,
			"%s: option %s must be followed by an argument\n",
			progname, **ppp
		    );
		}
		fprintf(stderr,
		    "%s -help gives usage information.\n",
		    progname
		);
		(void) exit(1);
	    }
	    switch (optp->argumentCount) {
	    case 0:
		result = (*(optp->handler))(
		    optp, **ppp, 0, 0
		);
		break;
	    case 1:
		result = (*(optp->handler))(
		    optp, **ppp, (*ppp)[1], 0
		);
		--*argcp; ++*ppp;
		break;
	    case 2:
		result = (*(optp->handler))(
		    optp, **ppp, (*ppp)[1], (*ppp)[2]
		);
		--*argcp; ++*ppp;
		--*argcp; ++*ppp;
		break;
	    default:
		fprintf(stderr,
		    "%s: %s; %d; fatal internal error: %d arguments\n",
		    progname, __FILE__, __LINE__, optp->argumentCount
		);
		exit(1);
	    }

	    if (result < 0) {
		fprintf(stderr,
		    "%s -help gives usage information.\n",
		    progname
		);
		(void) exit(1);
	    }

	    return result;
	}
    }
    return -1;
}

extern char *defaultPattern;

void
fListOptions(f)
    FILE *f;
{
    t_option *optp;

    for (optp = options; optp->names; optp++) {
	int i;

	if (optp->flags & OPTFLAG_UNDOCUMENTED) {
	    continue;
	}

	fprintf(f, "%s", optp->names);
	for (i = 0; i < optp->argumentCount; i++) {
	    fprintf(f, " x");
	}
	fprintf(f, "\n    %s\n", optp->description);
	if (optp->defaultValue) {
	    fprintf(f, "    [default: \"%s\"]\n", optp->defaultValue);
	}
    }
}

int
parseOptions(iargc, argv)
    int iargc;
    char *argv[];
{
    char **pp = argv;
    int argc;

    defaultPattern = (char *) DEFAULT_CONSTRAINT;
    SetConstraintFromString(defaultPattern);

    for (argc = iargc; argc > 1; argc--) {
	++pp;

	if (**pp != '-') {
	    /* it's not an option -- but we don't take any other
	     * kind of argument:
	     */
	    fprintf(stderr,
		"%s: unknown argument \"%s\"; -help gives usage\n",
		progname,
		argv[0]
	    );
	    return -1;
	}

	if (pp[0][1] == '-' && pp[0][2] == '\0') {
	    /* end of arguments */
	    return argc - iargc;
	}

	if (handleOneOption(&argc, &pp) < 0) {
	    fprintf(stderr,
		"%s: unknown option \"%s\"; -help gives usage\n",
		argv[0],
		*pp
	    );
	    return -1;
	}
    }
    defaultPattern = updateConstraintPattern();
    return 0;
}

